import { formatProposalMessage } from "./announcements";
import axios from "axios";
import { apiUrl } from "../../config";

//types

import { ApiPromise } from "@polkadot/api";
import { Council, ProposalArray, ProposalDetail } from "../types";
import { AccountId } from "@polkadot/types/interfaces";
import { getMember, getMemberIdByAccount } from "../joystream-lib/api";
import { MemberHandles, ProposalVotes } from "../joystream-lib/types";
import {
  ChannelId,
  PostId,
  ProposalDetailsOf,
  ThreadId,
} from "@joystream/types/augment";
import { Category, CategoryId } from "@joystream/types/forum";
import { MemberId, Membership } from "@joystream/types/members";
import { Proposal } from "@joystream/types/proposals";

// api

export const timestamp = async (api: ApiPromise) =>
  (await api.query.timestamp.now()).toNumber();

export const memberHandle = async (
  api: ApiPromise,
  id: MemberId
): Promise<string> => {
  const membership = (await api.query.members.membershipById(id)) as Membership;
  return membership.handle.toJSON();
};

export const memberHandleByAccount = async (
  api: ApiPromise,
  account: string
): Promise<string> => {
  const id = (await api.query.members.memberIdsByRootAccountId(
    account
  )) as MemberId;
  const handle: string = await memberHandle(api, id);
  return handle;
};

// forum

export const categoryById = async (
  api: ApiPromise,
  id: number
): Promise<Category> => {
  const category = (await api.query.forum.categoryById(id)) as Category;
  return category;
};

export const currentPostId = async (api: ApiPromise): Promise<number> => {
  const postId = (await api.query.forum.nextPostId()) as PostId;
  return postId.toNumber() - 1;
};

export const currentThreadId = async (api: ApiPromise): Promise<number> => {
  const threadId = (await api.query.forum.nextThreadId()) as ThreadId;
  return threadId.toNumber() - 1;
};

export const currentCategoryId = async (api: ApiPromise): Promise<number> => {
  const categoryId = (await api.query.forum.nextCategoryId()) as CategoryId;
  return categoryId.toNumber() - 1;
};

// proposals

export const proposalCount = async (api: ApiPromise): Promise<number> =>
  Number(await api.query.proposalsEngine.proposalCount());

export const activeProposals = async (
  api: ApiPromise,
  last: number
): Promise<number[]> => {
  let count = 0;
  try {
    console.log(`fetching active proposal count`);
    count = Number(await api.query.proposalsEngine.activeProposalCount());
  } catch (e) {
    console.error(`failed to fetch active proposal count: ${e.message}`);
  }
  let ids: number[] = [];
  for (let id = last; ids.length < count; id--) {
    if ([837, 839].includes(id)) continue;
    try {
      console.log(`fetching proposal ${id}`);
      const proposal = await proposalDetail(api, id);
      if (proposal && proposal.result === "Pending") ids.push(id);
    } catch (e) {
      console.error(`Failed to fetch proposal ${id}: ${e.message}`);
    }
  }
  return ids;
};

const getProposalType = async (
  api: ApiPromise,
  id: number
): Promise<string> => {
  const details = (await api.query.proposalsCodex.proposalDetailsByProposalId(
    id
  )) as ProposalDetailsOf;
  const [type]: string[] = Object.getOwnPropertyNames(details.toJSON());
  return type;
};

export const proposalDetail = async (
  api: ApiPromise,
  id: number
): Promise<ProposalDetail | undefined> => {
  let proposal: Proposal;
  try {
    proposal = (await api.query.proposalsEngine.proposals(id)) as Proposal;
    if (!proposal) return;
  } catch (e) {
    console.log(`Failed to fetch proposal detail ${id}`);
    return;
  }
  const status: { [key: string]: any } = proposal.status;
  if (!status) return;
  const stage: string = status.isActive ? "Active" : "Finalized";
  const { finalizedAt, proposalStatus } = status[`as${stage}`];
  const result: string = proposalStatus
    ? (proposalStatus.isApproved && "Approved") ||
      (proposalStatus.isCanceled && "Canceled") ||
      (proposalStatus.isExpired && "Expired") ||
      (proposalStatus.isRejected && "Rejected") ||
      (proposalStatus.isSlashed && "Slashed") ||
      (proposalStatus.isVetoed && "Vetoed")
    : "Pending";
  const { parameters, proposerId } = proposal;
  const author: string = await memberHandle(api, proposerId);
  const title: string = proposal.title.toString();
  const type: string = await getProposalType(api, id);
  const args: string[] = [String(id), title, type, stage, result, author];
  const message: { tg: string; discord: string } = formatProposalMessage(args);
  const createdAt: number = proposal.createdAt.toNumber();
  return { id, createdAt, finalizedAt, parameters, message, stage, result };
};

// status endpoint

export const fetchTokenValue = async (): Promise<string> =>
  axios
    .get("https://status.joystream.org/status")
    .then(({ data }) => `${Math.floor(+data.price * 100000000) / 100} $`)
    .catch((e) => {
      console.log(`Failed to fetch status.`);
      return `?`;
    });

// hdyra

export const fetchStorageSize = async () => {
  // `https://status.joystream.org/status`
  const url = `https://joystreamstats.live/static/status.json`;
  const { data } = await axios.get(url);
  if (!data) return `?`; // TODO fall back to JSG status endpoint
  const GB = Math.round(data.media.size / 1000000000);
  return `${GB} GB`;
};

// jsstats

export const discordHandles = async (): Promise<MemberHandles[]> => {
  console.debug(`Fetching user handles`);
  const rawUrl = `https://raw.githubusercontent.com/Joystream/community-repo/master/council/council_member_discord_usernames.md`;
  return await axios
    .get(rawUrl)
    .then(({ data }) => {
      const rows = data
        .split(`\n`)
        .filter((line: string) => line.includes(`|`))
        .slice(2);
      return rows.map((row: string) => {
        const [, handle, discord, id, telegram, github] = row
          .split(`|`)
          .map((s) => s.trim());
        return { handle, discord: { handle: discord, id }, telegram, github };
      });
    })
    .catch((error) => console.error(`Fetch user handles.`, error.message));
};

export const consulHandles = async (
  api: ApiPromise,
  handles: MemberHandles[],
  account: AccountId
): Promise<MemberHandles> => {
  const memberId = await getMemberIdByAccount(api, account);
  const handle = await memberHandle(api, memberId);
  const consul = handles.find((m) => m.handle === handle) || handles[0];
  consul.handle = handle;
  return consul;
};

export const activeProposalsVotes = async (): Promise<ProposalVotes[]> => {
  console.debug(`Fetching active proposals`);
  const { data } = await axios.get(apiUrl + `/v1/proposals`);
  return data.filter((p: ProposalVotes) => p.result === "Pending");
};
